import time
import uuid
import logging
from threading import Thread
from multiprocessing import Process, Event


class _Consumer:
    """Consumer class."""
    def __init__(self, func, kwargs, queue, pipe):
        """Consumer superclass, contains all of the functionality. Calculon does not deal with this
        class directly, rather it instantiates ConsumerThread or ConsumerProcess that inherit from
        _Consumer and from either Thread or Process classes.

        **Keyword arguments**

        * queue -- instance of multiprocessing.Queue;
        * func -- function that processes values received from the queue, once at a time;
        * kwargs -- a dictionary of arguments passed to func;
        * pipe -- end of a multiprocessing.Pipe() to which consumer can write the results.
        """

        self.name = uuid.uuid1().hex
        self.func = func
        self.kwargs = kwargs if kwargs else {}
        self.queue = queue
        self.pipe = pipe

        # Exit event indicating that there will be no items out on the queue
        # anymore and that it's safe to exit once queue is empty.
        self.exit = Event()

    def run(self):
        """Runs the producer function once.

        **Keyword arguments**

            None

        Consumer will repeteadly try to get a value from the queue until all producers have
        shutdown and the queue is not empty. Each time it succeeds, the value is passed to
        the consumer function to process the value. Once there are no more values in the
        queue and all of the producers have stopped, each of the consumers will be called
        once more, to allow to perform any sort of cleanup that might be required.

        Along with each call to the consumer function, the following values are passed in
        the argument dictionary.

        * _name -- unique name of the consumer (uuid);
        * _value -- value from the queue to process during this call;
        * _last_call -- a flag that when set to `True` indicates that this is the last "cleanup" call to the consumer. Also note that if If `_last_call` is `True`, `_value` is `None`.
        * _result -- contains the return value of the previous call to the consumer function. Set to None on the first call.
        """

        self._result = None
        self.kwargs["_result"] = None

        try:
            while not (self.exit.is_set() and self.queue.qsize() == 0):
                # Value to process.
                value = None

                # We use a non-blocking call to avoid deadlocking when
                # the queue is empty.
                try:
                    value = self.queue.get(block=False)
                except:
                    time.sleep(1)
                    continue

                if value is not None:

                        # Special arguments get refreshed on every call.
                        self.kwargs["_name"] = self.name
                        self.kwargs["_value"] = value
                        self.kwargs["_last_call"] = False
                        self.kwargs["_result"] = self.func(self.kwargs)

            # Last call to the consumer.
            self.kwargs["_name"] = self.name
            self.kwargs["_value"] = None
            self.kwargs["_last_call"] = True
            self._result = self.func(self.kwargs)

            # Set result dictionary.
            self.result = {
                'name': self.name,
                'result': self._result
            }

        except Exception as e:
            self.result = {
                'name': self.name,
                'exception': e
            }

        # For multiprocessing we need to communicate results
        # through a pipe.
        if self.pipe:
            self.pipe.send(self.result)
            self.pipe.close()

    def shutdown(self):
        """Sets the exit flag. The exit flag is used by the `run()` code
        to check if the consumer can be shutdown safely. It is called from
        the Calculon instance and indicates that all of the producers have
        stopped running."""
        self.exit.set()


class ConsumerProcess(_Consumer, Process):
    """Instantiates _Consumer and Process superclasses."""
    def __init__(self, func, kwargs, queue, pipe):
        Process.__init__(self)
        _Consumer.__init__(self, func, kwargs, queue, pipe)


class ConsumerThread(_Consumer, Thread):
    """Instantiates _Consumer and Thread superclasses."""
    def __init__(self, func, kwargs, queue):
        Thread.__init__(self)
        _Consumer.__init__(self, func, kwargs, queue, None)



import random
import unittest
from multiprocessing import Queue, Pipe

from calculon import Calculon
from calculon import ProducerProcess, ProducerThread
from calculon import ConsumerProcess, ConsumerThread

NUM_RESULTS = 5


def prod_function(kwargs):
    """Producer function that generates random integers.
    It returns a list of generated integers plus an
    integer passed to it vis kwargs['add']"""

    # Check the name is available.
    kwargs['_name']

    queue = kwargs['_queue']
    add = kwargs['add']

    values = []
    for i in range(0, NUM_RESULTS):
        value = random.randint(1, 10)
        values.append(value)
        queue.put(value)

    values.append(int(add))

    return values


def cons_function(kwargs):
    """Consumer function gets integers from the queue and
    stores them in a list passed throughout the calls.
    At the last call, the list plus the kwargs['add'] is
    returned."""

    kwargs['_name']
    is_last_call = kwargs['_last_call']
    add = kwargs['add']
    value = kwargs['_value']
    result = kwargs['_result']

    # If our first call was our last call.
    if result is None and is_last_call:
        return [add]

    # Otherwise, if it's not our last call.
    if not is_last_call:
        if result is None:
            int_list = [value]
        else:
            int_list = result + [value]
    # And if it is.
    else:
        int_list = result + [int(add)]

    return int_list


class TestCalculon(unittest.TestCase):

    def setUp(self):
        self.queue = Queue()
        self.pipe = Pipe()

    def test_producer_thread(self):
        """Check producer by itself, running as a thread."""
        pt = ProducerThread(prod_function, {'add': 5}, self.queue)
        pt.start()
        pt.join()

        # The queue is the correct size.
        self.assertTrue(self.queue.qsize() == NUM_RESULTS)

        # All of the results are in the range.
        while self.queue.qsize() > 0:
            result = self.queue.get()
            self.assertTrue(result in range(1, 11))

        # Producer's range is NUM_RESULTS + 1 (add variable).
        self.assertTrue(len(pt.result["result"]) == NUM_RESULTS + 1)

    def test_producer_process(self):
        """Check producer by itself, running as a process."""
        pp = ProducerProcess(prod_function, {'add': 5}, self.queue, self.pipe[0])
        pp.start()
        pp.join()

        # The queue is the correct size.
        self.assertTrue(self.queue.qsize() == NUM_RESULTS)

        # All of the results are in the range.
        while self.queue.qsize() > 0:
            result = self.queue.get()
            self.assertTrue(result in range(1, 11))

        # Producer's range is NUM_RESULTS + 1 (add variable).
        self.assertTrue(len(self.pipe[1].recv()["result"]) == NUM_RESULTS + 1)

    def test_consumer_thread(self):
        """Check consumer by itself, running as a thread."""
        for i in range(0, NUM_RESULTS):
            self.queue.put(i)

        ct = ConsumerThread(cons_function, {'add': 5}, self.queue)
        ct.start()
        ct.shutdown()
        ct.join()

        # The queue is empty, the results contain all numbers.
        self.assertTrue(self.queue.qsize() == 0)
        self.assertTrue(len(ct.result["result"]) == NUM_RESULTS + 1)

        # All result values are in correct range.
        for value in ct.result["result"]:
            self.assertTrue(str(value), value in range(1, 11))

    def test_consumer_process(self):
        """Check consumer by itself, running as a process."""
        for i in range(0, NUM_RESULTS):
            self.queue.put(i)

        cp = ConsumerProcess(cons_function, {'add': 5}, self.queue, self.pipe[0])
        cp.start()
        cp.shutdown()
        cp.join()

        result = self.pipe[1].recv()["result"]

        # The queue is empty, the results contain all numbers.
        self.assertTrue(self.queue.qsize() == 0)
        self.assertTrue(len(result) == NUM_RESULTS + 1)

        # All result values are in correct range.
        for value in result:
            self.assertTrue(str(value), value in range(1, 11))

    def test_calculon_thread(self):
        """The four Calculon tests check that we can run calculon as thread/process/combination.
        The idea is to generate random numbers in producer, retrieve them through consumer,
        add up the result and compare the two."""
        NUM_PROD = 50
        NUM_CONS = 30

        p_args = []
        c_args = []

        for i in range(NUM_PROD):
            p_args.append({"add": 0})

        for i in range(NUM_CONS):
            c_args.append({"add": 0})

        c = Calculon(prod_function, p_args, True, cons_function, c_args, True)
        result = c.start()

        prod_sum = 0
        cons_sum = 0

        for p in result["producers"]:
            prod_sum += sum(p["result"])

        for c in result["consumers"]:
            cons_sum += sum(c["result"])

        self.assertTrue(prod_sum == cons_sum)

    def test_calculon_process(self):
        NUM_PROD = 50
        NUM_CONS = 30

        p_args = []
        c_args = []

        for i in range(NUM_PROD):
            p_args.append({"add": 0})

        for i in range(NUM_CONS):
            c_args.append({"add": 0})

        c = Calculon(prod_function, p_args, False, cons_function, c_args, False)
        result = c.start()

        prod_sum = 0
        cons_sum = 0

        for p in result["producers"]:
            prod_sum += sum(p["result"])

        for c in result["consumers"]:
            cons_sum += sum(c["result"])

        self.assertTrue(prod_sum == cons_sum)

    def test_calculon_hybrid_1(self):
        NUM_PROD = 50
        NUM_CONS = 30

        p_args = []
        c_args = []

        for i in range(NUM_PROD):
            p_args.append({"add": 0})

        for i in range(NUM_CONS):
            c_args.append({"add": 0})

        c = Calculon(prod_function, p_args, True, cons_function, c_args, False)
        result = c.start()

        prod_sum = 0
        cons_sum = 0

        for p in result["producers"]:
            prod_sum += sum(p["result"])

        for c in result["consumers"]:
            cons_sum += sum(c["result"])

        self.assertTrue(prod_sum == cons_sum)

    def test_calculon_hybrid_2(self):
        NUM_PROD = 50
        NUM_CONS = 30

        p_args = []
        c_args = []

        for i in range(NUM_PROD):
            p_args.append({"add": 0})

        for i in range(NUM_CONS):
            c_args.append({"add": 0})

        c = Calculon(prod_function, p_args, False, cons_function, c_args, True)
        result = c.start()

        prod_sum = 0
        cons_sum = 0

        for p in result["producers"]:
            prod_sum += sum(p["result"])

        for c in result["consumers"]:
            cons_sum += sum(c["result"])

        self.assertTrue(prod_sum == cons_sum)

    def test_exceptions(self):
        """Checks that exceptions don't kill Calculon"""
        NUM_PROD = 1
        NUM_CONS = 1

        p_args = []
        c_args = []

        for i in range(NUM_PROD):
            p_args.append({"add": None})

        for i in range(NUM_CONS):
            c_args.append({"add": None})

        c = Calculon(prod_function, p_args, False, cons_function, c_args, True)
        result = c.start()

if __name__ == '__main__':
    unittest.main()



import uuid
import logging
from threading import Thread
from multiprocessing import Process

class _Producer:

    """Producer class."""
    def __init__(self, func, kwargs, queue, pipe):
        """Producer superclass, contains all of the functionality. Calculon does not deal with this
        class directly, rather it instantiates ProducerThread or ProducerProcess that inherit from
        _Producer and from either Thread or Process classes.

        **Keyword arguments**

        * queue -- instance of multiprocessing.Queue;
        * func -- function that puts values into the queue, optional return value;
        * kwargs -- a dictionary of arguments passed to func;
        * pipe -- end of a multiprocessing.Pipe() to which producer can write the results.
        """

        self.name = uuid.uuid1().hex
        self.func = func
        self.kwargs = kwargs if kwargs else {}
        self.queue = queue
        self.pipe = pipe

    def run(self):
        """Runs the producer function once.

        **Keyword arguments**

            None

        When the producer function is called, two additional arguments are passed to it:

        * _name -- unique name of the producer (uuid);
        * _queue -- the queue object where to put the results;.

        **Returns**

        * If the run completed successfully, the method returns a dictionary with two keys:

            * "name" -- name of this producer (uuid);
            * "result" -- the return value returned from the producer function.

        * If the run completed unsuccessfully, the method returns a dictionary with two keys:

            * "name" -- name of this producer (uuid);
            * "exception" -- the exception object for the raised exception.
        """
        # Add two additional arguments.
        self.kwargs["_name"] = self.name
        self.kwargs["_queue"] = self.queue

        try:
            self._result = self.func(self.kwargs)

            # Set result dictionary.
            self.result = {
                'name': self.name,
                'result': self._result
            }
        except Exception as e:
            self.result = {
                'name': self.name,
                'exception': e
            }

        # For multiprocessing we need to communicate results through a pipe.
        if self.pipe:
            self.pipe.send(self.result)
            self.pipe.close()


class ProducerProcess(_Producer, Process):
    """Process-based producer class."""
    def __init__(self, func, kwargs, queue, pipe):
        """Instantiates _Producer and Process superclasses."""
        Process.__init__(self)
        _Producer.__init__(self, func, kwargs, queue, pipe)


class ProducerThread(_Producer, Thread):
    """Thread-based producer class."""
    def __init__(self, func, kwargs, queue):
        """Instantiates _Producer and Thread superclasses."""
        Thread.__init__(self)
        _Producer.__init__(self, func, kwargs, queue, None)



#!/usr/bin/env python
import os
from setuptools import setup


# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='calculon',
    version='1.1.1',
    author='Nikita Pchelin',
    author_email='nikita@pchelin.ca',
    url='https://github.com/jango/calculon',
    description=r"Implementation of the producer-consumer pattern"
                "with customizable producer and consumer methods and"
                "an ability to choose thread/process model for execution.",
    long_description=read('README.md'),
    classifiers=['Classifier: Development Status :: 5 - Production/Stable',
                 'Classifier: License :: OSI Approved :: MIT License',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Natural Language :: English',
                 'Operating System :: POSIX',
                 'Operating System :: Microsoft :: Windows',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: System :: Distributed Computing'],
    test_suite='calculon.test',
    packages=['calculon', 'calculon.test'],
)



from multiprocessing import Queue, Pipe

from Consumer import ConsumerProcess, ConsumerThread
from Producer import ProducerProcess, ProducerThread


class Calculon:
    """Producer-consumer class. Responsible for initializing producers and consumers and controlling execution."""
    def __init__(self, prod_func, prod_kwargs, prod_use_threads, cons_func, cons_kwargs, cons_use_threads):
        """Initializes Calculon.

        **Keyword arguments**

        * prod_func    -- producer function that accepts one argument (dictionary of values);
        * prod_kwargs  -- a list of dictionaries, each representing a set of arguments for an instance of the producer function;
        * prod_use_threads -- a flag specifying if threads are used to run producer code (if False, processes are used);
        * cons_func    -- consumer function that accepts one argument (dictionary of values)
        * cons_kwargs  -- a list of dictionaries, each representing a set of arguments for an instance of the consumer function
        * cons_use_threads -- a flag specifying if threads are used to run consumer code (if False, processes are used).
        """

        self.cons_func = cons_func
        self.cons_kwargs = cons_kwargs
        self.cons_use_threads = cons_use_threads

        self.prod_func = prod_func
        self.prod_kwargs = prod_kwargs
        self.prod_use_threads = prod_use_threads

        self.queue = Queue()

    def start(self):
        """Starts producer and consumer threads / processes and controls the execution.

        **Keyword arguments**

            None

        **Returns**
            Returns a dictionary containing two elements:

            * value for key "producers" contains a list of results returned by each of the producer instance.
            * value for key "consumers" contains a list of results returned by each of the consumer instance.
        """

        # Producers.
        prod_objs = []
        prod_pipes = []

        for id in range(0, len(self.prod_kwargs)):
            args = self.prod_kwargs[id]

            if self.prod_use_threads:
                prod_obj = ProducerThread(self.prod_func, args, self.queue)
            else:
                prod_pipes.append(Pipe())
                prod_obj = ProducerProcess(self.prod_func, args, self.queue, prod_pipes[id][0])

            prod_objs.append(prod_obj)
            prod_obj.start()

        # Consumers.
        cons_objs = []
        cons_pipes = []

        for id in range(0, len(self.cons_kwargs)):
            args = self.cons_kwargs[id]

            if self.cons_use_threads:
                cons_obj = ConsumerThread(self.cons_func, args, self.queue)
            else:
                cons_pipes.append(Pipe())
                cons_obj = ConsumerProcess(self.cons_func, args, self.queue, cons_pipes[id][0])

            cons_objs.append(cons_obj)
            cons_obj.start()

        # Join on the producers.
        for prod_obj in prod_objs:
            prod_obj.join()

        # Shut down  theconsumers.
        for cons_obj in cons_objs:
            cons_obj.shutdown()

        # Join on the consumers.
        for cons_obj in cons_objs:
            cons_obj.join()

        # Collect result values.
        result = {"producers": [],
                  "consumers": []}

        for counter, prod_obj in enumerate(prod_objs):
            if isinstance(prod_obj, ProducerProcess):
                res = prod_pipes[counter][1].recv()
            else:
                res = prod_obj.result

            result["producers"] = result["producers"] + [res]

        for counter, cons_obj in enumerate(cons_objs):
            if isinstance(cons_obj, ConsumerProcess):
                res = cons_pipes[counter][1].recv()
            else:
                res = cons_obj.result

            result["consumers"] = result["consumers"] + [res]

        return result
